#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>

#define TRUE  1
#define FALSE 0

// Typedefs
typedef int        boolean;

// Prototypes
void        countData(char *dir_str, boolean isRoot);
char       *toPath(char *dir, char *data);
void        usage(void);
void        extendedUsage(void);

// Variables
char           *progname;

boolean         dflag        = FALSE;
boolean         rflag        = FALSE;
boolean         lflag        = FALSE;
boolean         hflag        = FALSE;

unsigned long   reg_count    = 0;
unsigned long   dir_count    = 0;
unsigned long   chr_count    = 0;
unsigned long   blk_count    = 0;
unsigned long   lnk_count    = 0;
unsigned long   fifo_count   = 0;
unsigned long   sock_count   = 0;
unsigned long   unknow_count = 0;
unsigned long   error_count  = 0;
unsigned long   all_count    = 0;

/******************************************************************************
 * Main
 *****************************************************************************/
int main(int argc, char **argv) {
    char    ch;
    char   *dir_str;
    
    // program Name
    progname = argv[0];
    
    // parameter Init
    while ((ch = getopt(argc, argv, "rlhd:")) != -1) {
        switch(ch) {
            // recursive without link
            case 'r':
                rflag = TRUE;
                break;
                
            // recursive with link
            case 'l':
                lflag = TRUE;
                break;
                
            // help
            case 'h':
                hflag = TRUE;
                break;
                
            // directory
            case 'd':
                dflag   = TRUE;
                dir_str = optarg;
                break;
                
            default:
                usage();
        }
    }
    
    if (hflag == TRUE) {
        extendedUsage();
    }
    
    if (dflag == FALSE) {
        if ((dir_str = getenv("PWD")) == NULL) {
            printf("%s: Can't get Environment Variable PWD... exit!\n", progname);
            exit(0);
        }
    }
    
    countData(dir_str, TRUE);
    
    all_count = reg_count  + \
                dir_count  + \
                chr_count  + \
                blk_count  + \
                lnk_count  + \
                fifo_count + \
                sock_count + \
                unknow_count;
    
    printf("Regular File(s):     %lu\n", reg_count);
    printf("Directory(s):        %lu\n", dir_count);
    printf("Symbolic Link(s):    %lu\n", lnk_count);
    printf("Character Device(s): %lu\n", chr_count);
    printf("Block Device(s):     %lu\n", blk_count);
    printf("FIFO File(s):        %lu\n", fifo_count);
    printf("Socket File(s):      %lu\n", sock_count);
    printf("Unknow File(s):      %lu\n", unknow_count);
    printf("All Data:            %lu\n", all_count);
    printf("Error(s):            %lu\n", error_count);
    
    return 0;
}

/******************************************************************************
 * Directory Count
 *****************************************************************************/
void countData(char *dir_str, boolean isRoot) {
    DIR            *dir;
    struct dirent  *data;
    struct stat     data_stat;
    int             data_stat_ret;
    
    // open directory
    if ((dir = opendir(dir_str)) != NULL && errno == 0) {
        
        // read directory entry
        while ((data = readdir(dir)) != NULL) {
            
            // is the entry '.' or '..' -> continue
            if (strcmp(data->d_name, ".") == NULL || strcmp(data->d_name, "..") == NULL) {
                continue;
            }
            
            // with or without including symbolic links
            if (lflag == TRUE) {
                data_stat_ret = stat(toPath(dir_str, data->d_name), &data_stat);
            } else {
                data_stat_ret = lstat(toPath(dir_str, data->d_name), &data_stat);
            }
            
            // specifie the data
            if (data_stat_ret != NULL || errno != NULL) {
                printf("%s - %d\n", dir_str, errno);
                error_count++;
            // Directory
            } else if (S_ISDIR(data_stat.st_mode) == TRUE) {
                dir_count++;
                if (rflag == TRUE || lflag == TRUE) {
                    countData(toPath(dir_str, data->d_name), FALSE);
                }
            // Character Device
            } else if (S_ISCHR(data_stat.st_mode) == TRUE) {
                chr_count++;
            // Block Device
            } else if (S_ISBLK(data_stat.st_mode) == TRUE) {
                blk_count++;
            // Regular File
            } else if (S_ISREG(data_stat.st_mode) == TRUE) {
                reg_count++;
            // FIFO File
            } else if (S_ISFIFO(data_stat.st_mode) == TRUE) {
                fifo_count++;
            // Symbolic Link
            } else if (S_ISLNK(data_stat.st_mode) == TRUE) {
                lnk_count++;
            // Socket File
            } else if (S_ISSOCK(data_stat.st_mode) == TRUE) {
                sock_count++;
            // Unknow
            } else {
                unknow_count++;
            }
        }

    } else {
        if (isRoot == TRUE) {
            printf("Error Nr. %d (see /usr/include/errno.h)\n", errno);
        } else {
            printf("%s - %d\n", dir_str, errno);
            error_count++;
        }
    }
}
/******************************************************************************
 * Path = Directory + Data
 *****************************************************************************/
char *toPath(char *dir, char *data) {
    int         i = 0;
    int         k = 0;
    size_t      dir_len, data_len;
    char       *path;
    
    dir_len  = strlen(dir);
    data_len = strlen(data);
    
    path = (char *) malloc((dir_len + data_len + 2) * sizeof(char));
    
    // Directory
    for (i = 0; dir[i] != '\0'; i++) {
        path[k] = dir[i];
        k++;
    }
    
    // Last Directory Character != /
    if (path[k - 1] != '/') {
        path[k] = '/';
        k++;
    }
    
    // Data
    for (i = 0; data[i] != '\0'; i++) {
        path[k] = data[i];
        k++;
    }
    
    // End of String
    path[k] = '\0';
    
    return path;
}

/******************************************************************************
 * Usage
 *****************************************************************************/
void usage(void) {
    printf("%s [-r | -l] [-h] <dir>\n", progname);
    exit(0);
}

/******************************************************************************
 * Extended Usage
 *****************************************************************************/
void extendedUsage(void) {
    printf("%s [-r | -l] [-h] [-d <dir>]\n", progname);
    printf("    -d      directory to count\n");
    printf("    -r      recursive without symbolic link following\n");
    printf("    -l      recursive with symbolic link following\n");
    printf("    -h      help\n");
    exit(0);
}
